"""Markup span management """


from __future__ import (absolute_import,
                        division,
                        print_function,
                        unicode_literals)


__all__ = ['Span', 'Marker']


class Span(object):
    
    """Markup span information container """
    
    __slots__ = ['_start', '_end', '_stag', '_etag']
    
    def __init__(self, start, end, stag, etag):
        
        if not (isinstance(start, int) or start is None):
            raise ValueError('start must be integers or None, not {}'
                             .format(type(start).__name__))
        if not (isinstance(end, int) or end is None):
            raise ValueError('start must be integers or None, not {}'
                             .format(type(start).__name__))
        
        self._start = start
        self._end = end
        self._stag = stag
        self._etag = etag
    
    @property
    def start(self):
        return self._start
    
    @property
    def end(self):
        return self._end
    
    @property
    def stag(self):
        return self._stag
    
    @property
    def etag(self):
        return self._etag
    
    def __repr__(self):
        
        return ('{}.{}({}, {}, {!r}, {!r})'
                .format(__name__, self.__class__.__name__,
                        self.start, self.end, self.stag, self.etag))
    
    def __eq__(self, other):
        
        """return whether two span objects are the same """
        
        return (self.start == other.start and self.end == other.end
                and self.stag == other.stag and self.etag == other.etag)
    
    def __len__(self):
        
        """return size of the span """
        
        if not self.isfixed():
            raise ValueError('unfixed span does not have length')
        
        return max(self.end - self.start, 0)
    
    def copy(self, **kwds):
        
        """return copy of the span with new values specified """
        
        for key in kwds:
            if key not in ('start', 'end', 'stag', 'etag'):
                raise ValueError('invalid keyword argument {}'.format(key))
        
        start = kwds.get('start', self.start)
        end = kwds.get('end', self.end)
        stag = kwds.get('stag', self.stag)
        etag = kwds.get('etag', self.etag)
        return self.__class__(start, end, stag, etag)
    
    def isfixed(self):
        
        """return whether its length is fixed """
        
        return (self.start is not None
                and self.start >= 0
                and self.end is not None
                and self.end >= 0)
    
    def fix(self, length):
        
        """return span object with fixed length """
        
        if self.start is None:
            start = 0
        elif self.start < 0:
            start = max(length + self.start, 0)
        else:
            start = min(self.start, length)
        
        if self.end is None:
            end = length
        elif self.end < 0:
            end = max(length + self.end, 0)
        else:
            end = min(self.end, length)
        
        return self.__class__(start, end, self.stag, self.etag)
    
    def conflicts(self, other):
        
        """return whether the span range conflicts with the other """
        
        if not isinstance(other, Span):
            raise ValueError('argument must be a Span instance, not {}'
                             .format(type(other).__name__))
        if not self.isfixed():
            raise ValueError('checking span is not fixed')
        if not other.isfixed():
            raise ValueError('checked span is not fixed')
        
        return (self.start < other.start < self.end < other.end
                or other.start < self.start < other.end < self. end)


class Marker(object):
    
    """Manages text with markups """
    
    def __init__(self):
        
        self._fixed_spans = []
        self._unfixed_spans = []
    
    def markup(self, start, end, stag, etag):
        
        """Markup from `start` to `end` with `stag` and `etag` """
        
        span = Span(start, end, stag, etag)
        if span.isfixed():
            conflicting = find_conflict(self._fixed_spans, span)
            if conflicting is not None:
                raise ValueError('{!r} conflicts with existing {!r}'
                                 .format(span, conflicting))
            self._fixed_spans.append(span)
        else:
            self._unfixed_spans.append(span)
    
    def itertags(self, length):
        
        """iterate (index, stag/etag) tuples in appropriate order """
        
        nboundaries = length+1
        stags = [[] for __ in range(nboundaries)]
        etags = [[] for __ in range(nboundaries)]
        spans = []
        for orig_span in self._fixed_spans + self._unfixed_spans:
            span = orig_span.fix(length)
            if span.start >= length:
                continue
            if orig_span.isfixed():
                spans.append(span)
            else:
                conflicting = find_conflict(spans, span)
                if conflicting is not None:
                    raise ValueError('{!r} conflicts with existing {!r}'
                                     .format(span, conflicting))
                spans.append(span)
            # NOTE: The orders of tags in here are important because 
            # they affect default appearance orders of stags/etags 
            # even though they will be reordered bellow.
            stags[span.start].insert(0, span)   # the later the earlier
            etags[span.end].append(span)        # the later the later
        for i in range(nboundaries):
            stag_stack = stags[i]
            etag_stack = etags[i]
            stag_stack.sort(key=len, reverse=True)
            etag_stack.sort(key=len)
            later = []
            for span in stag_stack:
                if span:
                    # non-empty stags will be yielded later[*]
                    later.append(span)
                else:
                    # empty stags/etags
                    if span.stag is not None:
                        yield (i, span.stag)
                    if span.etag is not None:
                        yield (i, span.etag)
            for span in etag_stack:
                if span:
                    # non-empty etags
                    if span.etag is not None:
                        yield (i, span.etag)
            for span in later:
                # [*]non-empty stags
                if span.stag is not None:
                    yield (i, span.stag)
    
    def apply(self, seq):
        
        """iterate stags, etags and subsequences of seq in appropriate 
        order """
        
        j = 0
        for i, tag in self.itertags(len(seq)):
            if j < i:
                yield seq[j:i]
            yield tag
            j = i
        if j < len(seq):
            yield seq[j:]


def find_conflict(spans, span):
    
    """return the first conflicting Span object with span in spans
    
    returns None if there is no conflict between span and any items of spans
    """
    
    if not isinstance(span, Span):
        raise ValueError('second argument must be a Span object, not {}'
                         .format(type(span).__name__))
    
    for span2 in spans:
        if span.conflicts(span2):
            return span2
    return None

